[v3,2/3] Add an optional "alias" attribute to syscall entries.

Message ID 20181109200432.84491-3-jhb@FreeBSD.org
State New, archived
Headers

Commit Message

John Baldwin Nov. 9, 2018, 8:04 p.m. UTC
  When setting a syscall catchpoint by name, catch syscalls whose name
or alias matches the requested string.

When the ABI of a system call is changed in the FreeBSD kernel, this
is implemented by leaving a compatibility system call using the old
ABI at the existing "slot" and allocating a new system call for the
version using the new ABI.  For example, new fields were added to the
'struct kevent' used by the kevent() system call in FreeBSD 12.  The
previous kevent() system call in FreeBSD 12 kernels is now called
freebsd11_kevent() and is still used by older binaries compiled
against the older ABI.  The freebsd11_kevent() system call can be
tagged with an "alias" attribute of "kevent" permitting 'catch syscall
kevent' to catch both system calls and providing the expected user
behavior for both old and new binaries.  It also provides the expected
behavior if GDB is compiled on an older host (such as a FreeBSD 11
host).

gdb/ChangeLog:

	* NEWS: Add entry documenting system call aliases.
	* break-catch-syscall.c (catch_syscall_split_args): Pass 'result'
	to get_syscalls_by_name.
	* gdbarch.sh (UNKNOWN_SYSCALL): Remove.
	* gdbarch.h: Regenerate.
	* syscalls/gdb-syscalls.dtd (syscall): Add alias attribute.
	* xml-syscall.c [!HAVE_LIBEXPAT] (get_syscalls_by_name): Rename
	from get_syscall_by_name.  Now accepts a reference to a vector of
	integers and returns a bool.
	[HAVE_LIBEXPAT] (struct syscall_desc): Add alias member.
	(syscall_create_syscall_desc): Add alias parameter and pass it to
	syscall_desc constructor.
	(syscall_start_syscall): Handle alias attribute.
	(syscall_attr): Add alias attribute.
	(xml_get_syscalls_by_name): Rename from xml_get_syscall_number.
	Now accepts a reference to a vector of integers and returns a
	bool.  Add syscalls whose alias or name matches the requested
	name.
	(get_syscalls_by_name): Rename from get_syscall_by_name.  Now
	accepts a reference to a vector of integers and returns a bool.
	* xml-syscall.h (get_syscalls_by_name): Likewise.
---
 gdb/ChangeLog                 | 24 ++++++++++++++
 gdb/NEWS                      |  5 +++
 gdb/break-catch-syscall.c     | 11 ++-----
 gdb/gdbarch.h                 |  3 --
 gdb/gdbarch.sh                |  3 --
 gdb/syscalls/gdb-syscalls.dtd |  1 +
 gdb/xml-syscall.c             | 61 ++++++++++++++++++++---------------
 gdb/xml-syscall.h             | 10 +++---
 8 files changed, 74 insertions(+), 44 deletions(-)
  

Comments

Eli Zaretskii Nov. 9, 2018, 8:18 p.m. UTC | #1
> From: John Baldwin <jhb@FreeBSD.org>
> Date: Fri,  9 Nov 2018 12:04:31 -0800
> 
> diff --git a/gdb/NEWS b/gdb/NEWS
> index ff9b192a38..ff225f361e 100644
> --- a/gdb/NEWS
> +++ b/gdb/NEWS
> @@ -38,6 +38,11 @@
>    requires the use of a keyword.  Selecting a frame by level is
>    unchanged.  The MI comment "-stack-select-frame" is unchanged.
>  
> +* System call catchpoints now support system call aliases.  This is
> +  used with the FreeBSD system call list so that system call
> +  catchpoints will also trigger for legacy system call numbers used by
> +  older FreeBSD binaries.

IMO, this is very abstract and too vague to be useful.  Since we won't
have anything more detailed in the manual, would it be possible to
make this description more concrete?  E.g., how about using some of
the text and examples you used in your description of the issue:

> When setting a syscall catchpoint by name, catch syscalls whose name
> or alias matches the requested string.
> 
> When the ABI of a system call is changed in the FreeBSD kernel, this
> is implemented by leaving a compatibility system call using the old
> ABI at the existing "slot" and allocating a new system call for the
> version using the new ABI.  For example, new fields were added to the
> 'struct kevent' used by the kevent() system call in FreeBSD 12.  The
> previous kevent() system call in FreeBSD 12 kernels is now called
> freebsd11_kevent() and is still used by older binaries compiled
> against the older ABI.  The freebsd11_kevent() system call can be
> tagged with an "alias" attribute of "kevent" permitting 'catch syscall
> kevent' to catch both system calls and providing the expected user
> behavior for both old and new binaries.  It also provides the expected
> behavior if GDB is compiled on an older host (such as a FreeBSD 11
> host).

Thanks.
  
John Baldwin Nov. 16, 2018, 5:43 p.m. UTC | #2
On 11/9/18 12:18 PM, Eli Zaretskii wrote:
>> From: John Baldwin <jhb@FreeBSD.org>
>> Date: Fri,  9 Nov 2018 12:04:31 -0800
>>
>> diff --git a/gdb/NEWS b/gdb/NEWS
>> index ff9b192a38..ff225f361e 100644
>> --- a/gdb/NEWS
>> +++ b/gdb/NEWS
>> @@ -38,6 +38,11 @@
>>    requires the use of a keyword.  Selecting a frame by level is
>>    unchanged.  The MI comment "-stack-select-frame" is unchanged.
>>  
>> +* System call catchpoints now support system call aliases.  This is
>> +  used with the FreeBSD system call list so that system call
>> +  catchpoints will also trigger for legacy system call numbers used by
>> +  older FreeBSD binaries.
> 
> IMO, this is very abstract and too vague to be useful.  Since we won't
> have anything more detailed in the manual, would it be possible to
> make this description more concrete?  E.g., how about using some of
> the text and examples you used in your description of the issue:

Ok.  I wasn't too happy with my first attempt either.  I've tried to give
some more detail while not making it too long.  Let me know what you think:

* System call catchpoints now support system call aliases on FreeBSD.
  When the ABI of a system call changes in FreeBSD, this is
  implemented by leaving a compatibility system call using the old ABI
  at the existing number and allocating a new system call number for
  the new ABI.  For example, FreeBSD 12 altered the layout of 'struct
  kevent' used by the 'kevent' system call.  As a result, FreeBSD 12
  kernels ship with both 'kevent' and 'freebsd11_kevent' system calls.
  The 'freebsd11_kevent' system call is assigned an alias of 'kevent'
  so that a system call catchpoint for the 'kevent' system call will
  catch invocations of both the 'kevent' and 'freebsd11_kevent'
  binaries.  This ensures that 'kevent' system calls are caught for
  binaries using either the old or new ABIs.
  
Eli Zaretskii Nov. 16, 2018, 7:36 p.m. UTC | #3
> Cc: gdb-patches@sourceware.org
> From: John Baldwin <jhb@FreeBSD.org>
> Date: Fri, 16 Nov 2018 09:43:56 -0800
> 
> > IMO, this is very abstract and too vague to be useful.  Since we won't
> > have anything more detailed in the manual, would it be possible to
> > make this description more concrete?  E.g., how about using some of
> > the text and examples you used in your description of the issue:
> 
> Ok.  I wasn't too happy with my first attempt either.  I've tried to give
> some more detail while not making it too long.  Let me know what you think:
> 
> * System call catchpoints now support system call aliases on FreeBSD.
>   When the ABI of a system call changes in FreeBSD, this is
>   implemented by leaving a compatibility system call using the old ABI
>   at the existing number and allocating a new system call number for
>   the new ABI.  For example, FreeBSD 12 altered the layout of 'struct
>   kevent' used by the 'kevent' system call.  As a result, FreeBSD 12
>   kernels ship with both 'kevent' and 'freebsd11_kevent' system calls.
>   The 'freebsd11_kevent' system call is assigned an alias of 'kevent'
>   so that a system call catchpoint for the 'kevent' system call will
>   catch invocations of both the 'kevent' and 'freebsd11_kevent'
>   binaries.  This ensures that 'kevent' system calls are caught for
>   binaries using either the old or new ABIs.

LGTM, thanks.
  
Pedro Alves Nov. 27, 2018, 5:13 p.m. UTC | #4
On 11/09/2018 08:18 PM, Eli Zaretskii wrote:
>> From: John Baldwin <jhb@FreeBSD.org>
>> Date: Fri,  9 Nov 2018 12:04:31 -0800
>>
>> diff --git a/gdb/NEWS b/gdb/NEWS
>> index ff9b192a38..ff225f361e 100644
>> --- a/gdb/NEWS
>> +++ b/gdb/NEWS
>> @@ -38,6 +38,11 @@
>>    requires the use of a keyword.  Selecting a frame by level is
>>    unchanged.  The MI comment "-stack-select-frame" is unchanged.
>>  
>> +* System call catchpoints now support system call aliases.  This is
>> +  used with the FreeBSD system call list so that system call
>> +  catchpoints will also trigger for legacy system call numbers used by
>> +  older FreeBSD binaries.
> 
> IMO, this is very abstract and too vague to be useful.  Since we won't
> have anything more detailed in the manual, would it be possible to
> make this description more concrete?  E.g., how about using some of
> the text and examples you used in your description of the issue:

Should we, have something in the manual?

Thanks,
Pedro Alves
  

Patch

diff --git a/gdb/ChangeLog b/gdb/ChangeLog
index 81faa2994d..002ea3253c 100644
--- a/gdb/ChangeLog
+++ b/gdb/ChangeLog
@@ -1,3 +1,27 @@ 
+2018-11-09  John Baldwin  <jhb@FreeBSD.org>
+
+	* NEWS: Add entry documenting system call aliases.
+	* break-catch-syscall.c (catch_syscall_split_args): Pass 'result'
+	to get_syscalls_by_name.
+	* gdbarch.sh (UNKNOWN_SYSCALL): Remove.
+	* gdbarch.h: Regenerate.
+	* syscalls/gdb-syscalls.dtd (syscall): Add alias attribute.
+	* xml-syscall.c [!HAVE_LIBEXPAT] (get_syscalls_by_name): Rename
+	from get_syscall_by_name.  Now accepts a reference to a vector of
+	integers and returns a bool.
+	[HAVE_LIBEXPAT] (struct syscall_desc): Add alias member.
+	(syscall_create_syscall_desc): Add alias parameter and pass it to
+	syscall_desc constructor.
+	(syscall_start_syscall): Handle alias attribute.
+	(syscall_attr): Add alias attribute.
+	(xml_get_syscalls_by_name): Rename from xml_get_syscall_number.
+	Now accepts a reference to a vector of integers and returns a
+	bool.  Add syscalls whose alias or name matches the requested
+	name.
+	(get_syscalls_by_name): Rename from get_syscall_by_name.  Now
+	accepts a reference to a vector of integers and returns a bool.
+	* xml-syscall.h (get_syscalls_by_name): Likewise.
+
 2018-11-09  John Baldwin  <jhb@FreeBSD.org>
 
 	* break-catch-syscall.c (catch_syscall_split_args): Pass 'result'
diff --git a/gdb/NEWS b/gdb/NEWS
index ff9b192a38..ff225f361e 100644
--- a/gdb/NEWS
+++ b/gdb/NEWS
@@ -38,6 +38,11 @@ 
   requires the use of a keyword.  Selecting a frame by level is
   unchanged.  The MI comment "-stack-select-frame" is unchanged.
 
+* System call catchpoints now support system call aliases.  This is
+  used with the FreeBSD system call list so that system call
+  catchpoints will also trigger for legacy system call numbers used by
+  older FreeBSD binaries.
+
 * New targets
 
   NXP S12Z		s12z-*-elf
diff --git a/gdb/break-catch-syscall.c b/gdb/break-catch-syscall.c
index f3740a1673..61c667a261 100644
--- a/gdb/break-catch-syscall.c
+++ b/gdb/break-catch-syscall.c
@@ -419,18 +419,13 @@  catch_syscall_split_args (const char *arg)
 	}
       else
 	{
-	  /* We have a name.  Let's check if it's valid and convert it
-	     to a number.  */
-	  get_syscall_by_name (gdbarch, cur_name, &s);
-
-	  if (s.number == UNKNOWN_SYSCALL)
+	  /* We have a name.  Let's check if it's valid and fetch a
+	     list of matching numbers.  */
+	  if (!get_syscalls_by_name (gdbarch, cur_name, result))
 	    /* Here we have to issue an error instead of a warning,
 	       because GDB cannot do anything useful if there's no
 	       syscall number to be caught.  */
 	    error (_("Unknown syscall name '%s'."), cur_name);
-
-	  /* Ok, it's valid.  */
-	  result.push_back (s.number);
 	}
     }
 
diff --git a/gdb/gdbarch.h b/gdb/gdbarch.h
index 2cb6961083..8e354d2278 100644
--- a/gdb/gdbarch.h
+++ b/gdb/gdbarch.h
@@ -1569,9 +1569,6 @@  typedef ULONGEST (gdbarch_type_align_ftype) (struct gdbarch *gdbarch, struct typ
 extern ULONGEST gdbarch_type_align (struct gdbarch *gdbarch, struct type *type);
 extern void set_gdbarch_type_align (struct gdbarch *gdbarch, gdbarch_type_align_ftype *type_align);
 
-/* Definition for an unknown syscall, used basically in error-cases.  */
-#define UNKNOWN_SYSCALL (-1)
-
 extern struct gdbarch_tdep *gdbarch_tdep (struct gdbarch *gdbarch);
 
 
diff --git a/gdb/gdbarch.sh b/gdb/gdbarch.sh
index bbfa8d2205..ff0e751f48 100755
--- a/gdb/gdbarch.sh
+++ b/gdb/gdbarch.sh
@@ -1393,9 +1393,6 @@  done
 # close it off
 cat <<EOF
 
-/* Definition for an unknown syscall, used basically in error-cases.  */
-#define UNKNOWN_SYSCALL (-1)
-
 extern struct gdbarch_tdep *gdbarch_tdep (struct gdbarch *gdbarch);
 
 
diff --git a/gdb/syscalls/gdb-syscalls.dtd b/gdb/syscalls/gdb-syscalls.dtd
index c2aa478aa4..6aa73f288a 100644
--- a/gdb/syscalls/gdb-syscalls.dtd
+++ b/gdb/syscalls/gdb-syscalls.dtd
@@ -12,4 +12,5 @@ 
 <!ATTLIST syscall
 	name			CDATA	#REQUIRED
 	number			CDATA	#REQUIRED
+	alias			CDATA	#IMPLIED
 	groups			CDATA	#IMPLIED>
diff --git a/gdb/xml-syscall.c b/gdb/xml-syscall.c
index 664cee77dd..0e763045b4 100644
--- a/gdb/xml-syscall.c
+++ b/gdb/xml-syscall.c
@@ -61,13 +61,12 @@  get_syscall_by_number (struct gdbarch *gdbarch,
   s->name = NULL;
 }
 
-void
-get_syscall_by_name (struct gdbarch *gdbarch, const char *syscall_name,
-		     struct syscall *s)
+bool
+get_syscalls_by_name (struct gdbarch *gdbarch, const char *syscall_name,
+		      std::vector<int> &syscall_numbers)
 {
   syscall_warn_user ();
-  s->number = UNKNOWN_SYSCALL;
-  s->name = syscall_name;
+  return false;
 }
 
 const char **
@@ -97,8 +96,8 @@  get_syscall_group_names (struct gdbarch *gdbarch)
 /* Structure which describes a syscall.  */
 struct syscall_desc
 {
-  syscall_desc (int number_, std::string name_)
-  : number (number_), name (name_)
+  syscall_desc (int number_, std::string name_, std::string alias_)
+  : number (number_), name (name_), alias (alias_)
   {}
 
   /* The syscall number.  */
@@ -108,6 +107,10 @@  struct syscall_desc
   /* The syscall name.  */
 
   std::string name;
+
+  /* An optional alias.  */
+
+  std::string alias;
 };
 
 typedef std::unique_ptr<syscall_desc> syscall_desc_up;
@@ -207,10 +210,11 @@  syscall_group_add_syscall (struct syscalls_info *syscalls_info,
 
 static void
 syscall_create_syscall_desc (struct syscalls_info *syscalls_info,
-			     const char *name, int number,
+			     const char *name, int number, const char *alias,
 			     char *groups)
 {
-  syscall_desc *sysdesc = new syscall_desc (number, name);
+  syscall_desc *sysdesc = new syscall_desc (number, name,
+					    alias != NULL ? alias : "");
 
   syscalls_info->syscalls.emplace_back (sysdesc);
 
@@ -235,6 +239,7 @@  syscall_start_syscall (struct gdb_xml_parser *parser,
   /* syscall info.  */
   char *name = NULL;
   int number = 0;
+  char *alias = NULL;
   char *groups = NULL;
 
   for (const gdb_xml_value &attr : attributes)
@@ -243,6 +248,8 @@  syscall_start_syscall (struct gdb_xml_parser *parser,
         name = (char *) attr.value.get ();
       else if (strcmp (attr.name, "number") == 0)
         number = * (ULONGEST *) attr.value.get ();
+      else if (strcmp (attr.name, "alias") == 0)
+        alias = (char *) attr.value.get ();
       else if (strcmp (attr.name, "groups") == 0)
         groups = (char *) attr.value.get ();
       else
@@ -251,7 +258,8 @@  syscall_start_syscall (struct gdb_xml_parser *parser,
     }
 
   gdb_assert (name);
-  syscall_create_syscall_desc (data->syscalls_info, name, number, groups);
+  syscall_create_syscall_desc (data->syscalls_info, name, number, alias,
+			       groups);
 }
 
 
@@ -259,6 +267,7 @@  syscall_start_syscall (struct gdb_xml_parser *parser,
 static const struct gdb_xml_attribute syscall_attr[] = {
   { "number", GDB_XML_AF_NONE, gdb_xml_parse_attr_ulongest, NULL },
   { "name", GDB_XML_AF_NONE, NULL, NULL },
+  { "alias", GDB_XML_AF_OPTIONAL, NULL, NULL },
   { "groups", GDB_XML_AF_OPTIONAL, NULL, NULL },
   { NULL, GDB_XML_AF_NONE, NULL, NULL }
 };
@@ -390,21 +399,22 @@  syscall_group_get_group_by_name (const struct syscalls_info *syscalls_info,
   return NULL;
 }
 
-static int
-xml_get_syscall_number (struct gdbarch *gdbarch,
-                        const char *syscall_name)
+static bool
+xml_get_syscalls_by_name (struct gdbarch *gdbarch, const char *syscall_name,
+			  std::vector<int> &syscall_numbers)
 {
   struct syscalls_info *syscalls_info = gdbarch_syscalls_info (gdbarch);
 
-  if (syscalls_info == NULL
-      || syscall_name == NULL)
-    return UNKNOWN_SYSCALL;
+  bool found = false;
+  if (syscalls_info != NULL && syscall_name != NULL)
+    for (const syscall_desc_up &sysdesc : syscalls_info->syscalls)
+      if (sysdesc->name == syscall_name || sysdesc->alias == syscall_name)
+	{
+	  syscall_numbers.push_back (sysdesc->number);
+	  found = true;
+	}
 
-  for (const syscall_desc_up &sysdesc : syscalls_info->syscalls)
-    if (sysdesc->name == syscall_name)
-      return sysdesc->number;
-
-  return UNKNOWN_SYSCALL;
+  return found;
 }
 
 static const char *
@@ -510,14 +520,13 @@  get_syscall_by_number (struct gdbarch *gdbarch,
   s->name = xml_get_syscall_name (gdbarch, syscall_number);
 }
 
-void
-get_syscall_by_name (struct gdbarch *gdbarch,
-		     const char *syscall_name, struct syscall *s)
+bool
+get_syscalls_by_name (struct gdbarch *gdbarch, const char *syscall_name,
+		      std::vector<int> &syscall_numbers)
 {
   init_syscalls_info (gdbarch);
 
-  s->number = xml_get_syscall_number (gdbarch, syscall_name);
-  s->name = syscall_name;
+  return xml_get_syscalls_by_name (gdbarch, syscall_name, syscall_numbers);
 }
 
 const char **
diff --git a/gdb/xml-syscall.h b/gdb/xml-syscall.h
index 2e76a9a97f..a3b032aeb3 100644
--- a/gdb/xml-syscall.h
+++ b/gdb/xml-syscall.h
@@ -38,11 +38,13 @@  void set_xml_syscall_file_name (struct gdbarch *gdbarch,
 void get_syscall_by_number (struct gdbarch *gdbarch,
 			    int syscall_number, struct syscall *s);
 
-/* Function that retrieves the syscall number corresponding to the given
-   name.  It puts the requested information inside 'struct syscall'.  */
+/* Function that retrieves the syscall numbers corresponding to the
+   given name.  The numbers of all syscalls with either a name or
+   alias equal to SYSCALL_NAME are appended to SYSCALL_NUMBERS.  If no
+   matching syscalls are found, return false.  */
 
-void get_syscall_by_name (struct gdbarch *gdbarch,
-			  const char *syscall_name, struct syscall *s);
+bool get_syscalls_by_name (struct gdbarch *gdbarch, const char *syscall_name,
+			   std::vector<int> &syscall_numbers);
 
 /* Function used to retrieve the list of syscalls in the system.  This list
    is returned as an array of strings.  Returns the list of syscalls in the